@@ -159,17 +159,28 @@ struct handshake_state {
 	void *user_data;
 
 	void (*free)(struct handshake_state *s);
+	bool in_event;
 
 	handshake_event_func_t event_func;
 };
 
-#define handshake_event(hs, event, ...)	\
-	do {	\
-		if (!(hs)->event_func)	\
-			break;	\
+#define handshake_event(_hs, event, ...)	\
+	(__extension__ ({	\
+		bool freed = false;	\
 	\
-		(hs)->event_func((hs), event, (hs)->user_data, ##__VA_ARGS__); \
-	} while (0)
+		if ((_hs)->event_func && !(_hs)->in_event) {	\
+			(_hs)->in_event = true;	\
+			(_hs)->event_func((_hs), event, (_hs)->user_data,	\
+					##__VA_ARGS__); \
+	\
+			if (!(_hs)->in_event) {	\
+				handshake_state_free((_hs));	\
+				freed = true;	\
+			} else	\
+				(_hs)->in_event = false;	\
+		}	\
+		freed;	\
+	}))
 
 void handshake_state_free(struct handshake_state *s);
 
