@@ -101,4 +101,21 @@ static inline bool util_ip_subnet_match(uint8_t prefix_len,
 		  ~((1u << (8 - (prefix_len % 8))) - 1));
 }
 
+typedef void (*scan_freq_set_func_t)(uint32_t freq, void *userdata);
+
+struct scan_freq_set *scan_freq_set_new(void);
+void scan_freq_set_free(struct scan_freq_set *freqs);
+bool scan_freq_set_add(struct scan_freq_set *freqs, uint32_t freq);
+bool scan_freq_set_contains(const struct scan_freq_set *freqs, uint32_t freq);
+uint32_t scan_freq_set_get_bands(struct scan_freq_set *freqs);
+void scan_freq_set_foreach(const struct scan_freq_set *freqs,
+				scan_freq_set_func_t func, void *user_data);
+void scan_freq_set_merge(struct scan_freq_set *to,
+					const struct scan_freq_set *from);
+void scan_freq_set_constrain(struct scan_freq_set *set,
+					const struct scan_freq_set *constraint);
+bool scan_freq_set_isempty(const struct scan_freq_set *set);
+uint32_t *scan_freq_set_to_fixed_array(const struct scan_freq_set *set,
+					size_t *len_out);
+
 #endif /* __UTIL_H */
